/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <syslog.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>

#ifdef HAVE_SYSEXITS_H
# include <sysexits.h>
#else
# define EX_OK          0       /* successful termination */
# define EX__BASE       64      /* base value for error messages */
# define EX_USAGE       64      /* command line usage error */
# define EX_DATAERR     65      /* data format error */
# define EX_NOINPUT     66      /* cannot open input */
# define EX_NOUSER      67      /* addressee unknown */
# define EX_NOHOST      68      /* host name unknown */
# define EX_UNAVAILABLE 69      /* service unavailable */
# define EX_SOFTWARE    70      /* internal software error */
# define EX_OSERR       71      /* system error (e.g., can't fork) */
# define EX_OSFILE      72      /* critical OS file missing */
# define EX_CANTCREAT   73      /* can't create (user) output file */
# define EX_IOERR       74      /* input/output error */
# define EX_TEMPFAIL    75      /* temp failure; user is invited to retry */
# define EX_PROTOCOL    76      /* remote error in protocol */
# define EX_NOPERM      77      /* permission denied */
# define EX_CONFIG      78      /* configuration error */
# define EX__MAX        78      /* maximum listed value */
#endif

#include <mailutils/mailutils.h>
#include <mailutils/argp.h>
#include <mailutils/daemon.h>
#include <libmilter/mfapi.h>

#include "mailfrom.h"


/* Configurable options */

#define MAILFROMD_DAEMON        0
#define MAILFROMD_TEST          1
#define MAILFROMD_DELETE        2
#define MAILFROMD_DELETE_RATEDB 3
#define MAILFROMD_LISTDB        4
#define MAILFROMD_LIST_RATEDB   5
#define MAILFROMD_EXPIRE        6
#define MAILFROMD_EXPIRE_RATEDB 7
#define MAILFROMD_SHOW_DEFAULTS 8

int mode = MAILFROMD_DAEMON; /* Default operation mode */

int need_config = 1;/* Set if the current mode requires reading the
		       configuration file */
char *config_file = DEFAULT_CONFIG_FILE;
int config_ldebug;  /* Enable tracing the lexical analyzer */
int config_ydebug;  /* Enable tracing the parser */
int config_dump;    /* Dump created config tree to stdout */
int do_transcript;  /* Enable session transript */
int do_trace;       /* Enable tracing configuration */
int debug_level;    /* Debugging level */
int log_to_stderr;  /* Use stderr for logging */
char *smtp_domain;  /* Default SMTP domain for EHLO command */
char *portspec = DEFAULT_SOCKET; /* Communication socket specification */
int force_remove;   /* Remove local communication socket if it already
		       exists */
int foreground;     /* Stay in foreground */
char *postmaster_email = "";

char *pidfile = DEFAULT_PIDFILE;
char *user = DEFAULT_USER;       /* Switch to this user privileges after 
				    startup */
#ifdef USE_DBM
time_t positive_expire_interval = DEFAULT_EXPIRE_INTERVAL;
                                 /* Expire positive cache entries after this
				    number of seconds */
time_t negative_expire_interval = DEFAULT_EXPIRE_INTERVAL/2;
                                 /* Expire negative cache entries after this
				    number of seconds */
time_t rates_expire_interval = DEFAULT_EXPIRE_RATES_INTERVAL;
                                 /* Expire email rate entries after this
				    number of seconds */
int predict_next_option; 
double predict_rate;             /* Prediction rate for --list-rate */
#endif

/* I/O settings */
unsigned io_timeout = 3;
unsigned io_attempts = 3;


/* Per-message data */
struct smtp_io_data; /* Declared later on */
struct message_data {
	struct smtp_io_data *io;
	list_t hdr;
};

#define MLFIPRIV(ctx)  ((struct message_data*) smfi_getpriv(ctx))

static struct message_data *
priv_get(SMFICTX *ctx)
{
	struct message_data *md = MLFIPRIV(ctx);
	if (!md) {
		md = malloc(sizeof(*md));
		if (!md) 
			mu_error("not enough memory");
		else {
			md->hdr = NULL;
			md->io = NULL;
			smfi_setpriv(ctx, md);
		}
	}
	return md;
}

static int
priv_set_io(SMFICTX *ctx, struct smtp_io_data *io)
{
	struct message_data *md = priv_get(ctx);
	if (!md)
		return 1;
	md->io = io;
	return 0;
}

int
priv_store_header_command(SMFICTX *ctx, struct header_node *node)
{
	struct message_data *md = priv_get(ctx);
	if (!md)
		return 1;
	if (!md->hdr)
		list_create(&md->hdr);
	list_append(md->hdr, node);
	return 0;
}


/* Logging & debugging */

int
syslog_printer (int prio, const char *fmt, va_list ap)
{
#ifdef HAVE_VSYSLOG
  vsyslog (prio, fmt, ap);
#else
  char buf[128];
  vsnprintf (buf, sizeof buf, fmt, ap);
  syslog (prio, "%s", buf);
#endif
  return 0;
}

int
syslog_error_printer (const char *fmt, va_list ap)
{
	return syslog_printer(LOG_ERR, fmt, ap);
}

void
vlogmsg(int prio, const char *fmt, va_list ap)
{
	if (log_to_stderr) {
		vfprintf(stderr, fmt, ap);
		fprintf(stderr, "\n");
	} else
		syslog_printer(prio, fmt, ap);
}

void
logmsg(int prio, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(prio, fmt, ap);
	va_end(ap);
}

void
transcript(char *prefix, const char *msg)
{
	if (do_transcript) {
		int len = strlen(msg);
		if (msg[len-1] == '\n') {
			--len;
			if (len > 0 && msg[len-1] == '\r')
				--len;
		}
		if (len)
			logmsg(LOG_INFO, "%s %*.*s", prefix, len, len, msg);
	}
}

void
trace(const char *fmt, ...)
{
	if (do_trace) {
		va_list ap;
		va_start(ap, fmt);
		vlogmsg(LOG_INFO, fmt, ap);
		va_end(ap);
	}
}

void
debug_log(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(LOG_DEBUG, fmt, ap);
	va_end(ap);
}

void
log_status(sfsistat status, SMFICTX *ctx)
{
	if (debug_level >= 1) {
		char *id = smfi_getsymval(ctx, "i");
		const char *str = sfsistat_str(status);
		if (str)
			logmsg(LOG_INFO,
			       "%s: %s", id ? id : "(nil)", str);
		else
			logmsg(LOG_INFO,
			       "%s: status %d",
			       id ? id : "(nil)", status);
	}
}


/* Sendmail class file support */

static list_t domain_list;

static int
compare_string(const void *item, const void *value)
{
	return strcmp(item, value);
}

/* Read domains from sendmail-style domain file NAME and store them in
   DOMAIN_LIST */
void
read_domain_file(char *name)
{
	FILE *fp;
	char buf[256];
	char *p;
	
	fp = fopen(name, "r");
	if (!fp) {
		mu_error("cannot open file `%s': %s",
			 name, mu_strerror(errno));
		return;
	}

	if (!domain_list) {
		list_create(&domain_list);
		list_set_comparator(domain_list, compare_string);
	}

	while (p = fgets(buf, sizeof buf, fp)) {
		char *q;
		
		while (*p && isspace(*p))
			p++;
		if (*p == '#')
			continue;

		for (q = p + strlen(p) - 1; q > p && isspace(*q); q--)
			*q = 0;

		if (*p == 0)
			continue;
		
		list_append(domain_list, strdup(p));
	}

	fclose(fp);
}

/* Return true if we relay domain NAME */
int
relayed_domain_p(char *name)
{
	char *p = name;

	while (p) {
		if (list_locate(domain_list, p, NULL) == 0) {
			debug2(10,"%s is in relayed domain %s",
				  name, p);
			return 1;
		}
		p = strchr(p, '.');
		if (p)
			p++;
	}
	return 0;
}

int
resolve_ipstr(char *client, char *hbuf, size_t size)
{
	struct sockaddr_in sa;
	
	memset(&sa, 0, sizeof sa);
	sa.sin_family = AF_INET;
	if (inet_aton(client, &sa.sin_addr) == 0)
		return 1;
	return getnameinfo((struct sockaddr *)&sa, sizeof sa,
			   hbuf, size,
			   NULL, 0,
			   0);
}

/* Return true if CLIENT represents a host we relay. CLIENT is a dotted-quad
   IP address. */
int
host_in_relayed_domain_p(char *client)
{
	char hbuf[NI_MAXHOST];
	
	if (list_is_empty(domain_list))
		return 0;

	if (resolve_ipstr(client, hbuf, sizeof(hbuf)))
		return 0;
	return relayed_domain_p(hbuf);
}


/* SMTP I/O functions */

#define SMTP_MAJOR(c) ((c)/100)

struct smtp_io_data {
	SMFICTX *ctx;        /* Milter context */
	stream_t stream;     /* I/O stream */
	size_t send_off;     /* Send offset */
	size_t recv_off;     /* Receive offset */
	struct obstack stk;  /* Obstack for keeping commands/replies */
	char *command;       /* Last issued command */ 
	char *reply;         /* Last received reply */
	char *start;         /* First line of the reply, if it was multiline */
	char buf[128];       /* Input buffer */
	size_t level;        /* Number of bytes in buf */
	int code;            /* Reply code */
};

int
smtp_io_data_init(struct smtp_io_data *dat, SMFICTX *ctx, stream_t stream)
{
	obstack_init(&dat->stk);
	dat->send_off = dat->recv_off = 0;
	dat->stream = stream;
	dat->start = dat->command = dat->reply = NULL;
	dat->level = 0;
	dat->ctx = ctx;
	if (ctx && priv_set_io(ctx, dat)) /* save the private data */
		return 1;
	return 0;
}

void
smtp_io_data_destroy(struct smtp_io_data *dat)
{
	if (dat) {
		stream_close (dat->stream);
		stream_destroy(&dat->stream, stream_get_owner(dat->stream));
		obstack_free(&dat->stk, NULL);
		if (dat->ctx)
			priv_set_io(dat->ctx, NULL);
	}
}

int
smtp_stream_wait(stream_t stream, int flags, int *attempt)
{
	int rc;
	int oflags = flags;
	struct timeval tv;
	while (*attempt < io_attempts) {
		tv.tv_sec = io_timeout;
		tv.tv_usec = 0;
		rc = stream_wait(stream, &oflags, &tv);
		switch (rc) {
		case 0:
			if (flags & oflags)
				return 0;
			/* FALLTHROUGH */
		case EAGAIN:
		case EINPROGRESS:
			++*attempt;
			continue;

		default:
			return rc;
		}
	}
			
	return ETIMEDOUT;
}

int
smtp_wait(struct smtp_io_data *dat, int flags, int *attempt)
{
	return smtp_stream_wait(dat->stream, flags, attempt);
}

int
smtp_send(struct smtp_io_data *dat, char *command)
{
	int attempt = 0;
	size_t len = strlen(command);

	transcript("SEND:", command);
	do {
		size_t nb;
		int rc = stream_write(dat->stream, command, len,
				      dat->send_off, &nb);
		if (rc == 0) {
			if (nb == 0) {
				mu_error("stream_write: wrote 0 bytes");
				return -1;
			}
			dat->send_off += nb;
			len -= nb;
			command += nb;
		} else if (rc == EAGAIN) {
			if (attempt < io_attempts) { 
				rc = smtp_wait(dat, MU_STREAM_READY_WR,
                                               &attempt);
				if (rc) {
					mu_error ("smtp_wait failed: %s",
						  mu_strerror(rc));
					return -1;
				}
				continue;
			} else {
				mu_error ("stream_write timed out");
				return -1;
			}
		} else {
			mu_error ("stream_write: %s", mu_strerror (rc));
			return -1;
		}
	} while (len > 0);
	return 0;
}

int
smtp_send2(struct smtp_io_data *dat, char *command, char *arg)
{
	obstack_grow(&dat->stk, command, strlen(command));
	if (arg)
		obstack_grow(&dat->stk, arg, strlen(arg));
	obstack_grow(&dat->stk, "\r\n", 2);
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	return smtp_send(dat, dat->command);
}

int
smtp_send3(struct smtp_io_data *dat, char *command, char *arg1, char *arg2)
{
	obstack_grow(&dat->stk, command, strlen(command));
	obstack_grow(&dat->stk, arg1, strlen(arg1));
	obstack_grow(&dat->stk, arg2, strlen(arg2));
	obstack_grow(&dat->stk, "\r\n", 2);
	obstack_1grow(&dat->stk, 0);
	dat->command = obstack_finish(&dat->stk);
	
	return smtp_send(dat, dat->command);
}

int
smtp_recvline(struct smtp_io_data *dat)
{
	int attempt = 0;
	for (;;) {
		char *p;
		
		if (dat->level == 0) {
			int rc = stream_read(dat->stream,
					     dat->buf, sizeof dat->buf,
					     dat->recv_off, &dat->level);
			if (rc == 0) {
				if (dat->level == 0) {
					mu_error("stream_read: read 0 bytes");
					return -1;
				}
				dat->recv_off += dat->level;
			} else if (rc == EAGAIN) {
				if (attempt < io_attempts) {
					rc = smtp_wait(dat, 
						       MU_STREAM_READY_RD, 
						       &attempt);
					if (rc) {
						mu_error ("smtp_wait failed: %s",
                                                          mu_strerror(rc));
						return -1;
					}
					continue;
				} else {
					mu_error("stream_read timed out");
					return -1;
				}
			} else {
				mu_error ("stream_read: %s", mu_strerror (rc));
				return -1;
			}
		}
		attempt = 0;
		
		p = memchr(dat->buf, '\n', dat->level);
		if (!p) {
			obstack_grow(&dat->stk, dat->buf, dat->level);
			dat->level = 0;
			continue;
		} else {
			size_t len = p - dat->buf + 1;
			obstack_grow(&dat->stk, dat->buf, len);
			obstack_1grow(&dat->stk, 0);
			dat->reply = obstack_finish(&dat->stk);
			dat->level -= len;
			memmove(dat->buf, dat->buf + len, dat->level);
			break;
		}
	}
	return 0;
}

int
smtp_recv(struct smtp_io_data *dat)
{
	char *p;
	dat->start = NULL;
	do {
		int code;
		int rc = smtp_recvline(dat);
		if (rc)
			return -1;
		transcript("RECV:", dat->reply);
		code = strtoul (dat->reply, &p, 0);
		if (p - dat->reply != 3 || (*p != '-' && *p != ' ')) {
			mu_error("Unexpected reply from server: %s",
				 dat->reply);
			return -1;
		} else if (!dat->start) {
			dat->start = dat->reply;
			dat->code = code;
		} else if (dat->code != code) {
			mu_error("Unexpected reply code from server: %d",
				 code);
			return -1;
		}
	} while (*p == '-');
	obstack_1grow (&dat->stk, 0);
	obstack_finish (&dat->stk);
	return 0;
}


/* Milter-specific functions */

/* Verify whether EMAIL address is served by host CLIENT_ADDR. */
mf_status
check_on_host(SMFICTX *ctx, char *email, char *client_addr,
	      char *ehlo, char *mailfrom)
{
	int rc;
	stream_t stream;
	struct smtp_io_data io;
	mf_status status = mf_success;
	int count = 0;
	
	debug2(10, "email = %s, client_addr = %s", email, client_addr);
	rc = tcp_stream_create (&stream, client_addr, 25, MU_STREAM_NONBLOCK);
	if (rc) {
		mu_error ("tcp_stream_create: %s", mu_strerror (rc));
		return mf_temp_failure;
	}

	while ((rc = stream_open(stream))) {
		if ((rc == EAGAIN || rc == EINPROGRESS)
		    && count < io_attempts) {
			rc = smtp_stream_wait(stream, MU_STREAM_READY_WR,
					      &count);
			if (rc == 0)
				continue;
		}
		mu_error("stream_open: %s, %d", mu_strerror(rc), count);
		stream_destroy(&stream, stream_get_owner(stream));
		if (rc == EAGAIN || rc == EINPROGRESS)
			return mf_temp_failure;
		else
			return mf_failure;
	}
	debug(100,"stream opened");
	
	if (smtp_io_data_init(&io, ctx, stream))
		status = mf_temp_failure;
	else {
		do {
			if (smtp_recv(&io) || SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send2(&io, "HELO ", ehlo);
			if (smtp_recv(&io)) {
				status = mf_not_found;
				break;
			} else if (SMTP_MAJOR(io.code) == 5) {
				/* Let's try EHLO, then */
				smtp_send2(&io, "EHLO ", ehlo);
				if (smtp_recv(&io)) {
					status = mf_not_found;
					break;
				}
			}
			if (SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send3(&io, "MAIL FROM: <", mailfrom, ">");
			if (smtp_recv(&io) || SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send3(&io, "RCPT TO: <", email, ">");
			if (smtp_recv(&io) || SMTP_MAJOR(io.code) != 2) {
				status = SMTP_MAJOR(io.code) == 4 ?
					mf_temp_failure : mf_not_found;
				break;
			}
			smtp_send2(&io, "QUIT", NULL);
		} while (0);

		smtp_io_data_destroy(&io);
	}
	
	return status;
}

mf_status
check_mx_records(SMFICTX *ctx, char *email, char *client_addr,
		 char *ehlo, char *mailfrom)
{
	int i;
	mxbuf_t mxbuf;
	mf_status rc, mxstat = getmx(client_addr, mxbuf);

	switch (mxstat) {
	case mf_temp_failure:
		rc = mf_temp_failure;
		break;

	case mf_failure:
	case mf_not_found:
		rc = mf_not_found;
		break;

	case mf_success:
		debug1(2,"Checking MX servers for %s", email);
		rc = mf_not_found;
		for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++) {
			rc = check_on_host(ctx, email, mxbuf[i],
					   ehlo, mailfrom);
			if (mf_resolved(rc))
				break;
			free(mxbuf[i]);
		}
		for (; i < MAXMXCOUNT && mxbuf[i]; i++) 
			free(mxbuf[i]);
	}
	return rc;
}

/* Method "strict". Verifies whether EMAIL is understood either by
   host CLIENT_ADDR or one of MX servers of its domain */
mf_status
method_strict(SMFICTX *ctx, char *email, char *client_addr,
	      char *ehlo, char *mailfrom)
{
	mf_status rc = cache_get2(email, client_addr);

        if (email[0] == 0)
                return mf_success;

	if (!mf_resolved(rc)) {
		rc = check_on_host(ctx, email, client_addr, ehlo, mailfrom);
		if (!mf_resolved(rc)) {
			mf_status mx_stat;
			mx_stat = check_mx_records(ctx, email, client_addr,
						   ehlo, mailfrom);
			if (mf_resolved(mx_stat)
			    || mx_stat == mf_temp_failure)
				rc = mx_stat;
		}
		if (mf_resolved(rc))
			cache_insert2(email, client_addr, rc);
	}
	return rc;
}

/* Method "standard". Verifies whether EMAIL is understood by
   any of its MXs */
mf_status
method_standard(SMFICTX *ctx, char *email, char *ehlo, char *mailfrom)
{
	mf_status rc = cache_get(email);

        if (email[0] == 0)
                return mf_success;
	if (!mf_resolved(rc)) {	
		char *p = strchr(email, '@');
		if (p == NULL) {
			mu_error("Invalid address: %s", email);
			rc = mf_not_found;
		} else {
			rc = check_mx_records(ctx, email, p+1, ehlo, mailfrom);
			if (rc != mf_success) {
				mf_status host_stat;
				host_stat = check_on_host(ctx, email, p+1, 
							  ehlo, mailfrom);
				if (mf_resolved(host_stat)
				    || host_stat == mf_temp_failure)
					rc = host_stat;
			}
		}
		if (mf_resolved(rc))
			cache_insert(email, rc);
	}
	return rc;
}

/* Cleanup functions */
void
filter_cleanup(SMFICTX *ctx)
{
	struct message_data *md = MLFIPRIV(ctx);
	debug(100,"cleaning up");
	if (md) {
		smtp_io_data_destroy(md->io);
		list_destroy(&md->hdr);
		free(md);
		smfi_setpriv(ctx, NULL);
	}
}


/* Milter interface functions */

sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	sfsistat status;
	status = run_program(ctx);
	log_status(status, ctx);
	return status;
}

static int
run_header_command(void *item, void *data)
{
	struct header_node *hdr = item;
	SMFICTX *ctx = data;

	debug3(100,
	       "%s %s: %s",
	       header_command_str(hdr->opcode),
	       hdr->name, hdr->value);
	switch (hdr->opcode) {
	case header_add:
		smfi_addheader(ctx, hdr->name, hdr->value);
		break;
		
	case header_replace:
		smfi_chgheader(ctx, hdr->name, 1, hdr->value);
		break;

	case header_delete:
		smfi_chgheader(ctx, hdr->name, 1, NULL);
		break;
	}
	return 0;
}

	
sfsistat
mlfi_eom(SMFICTX *ctx)
{
	struct message_data *md = MLFIPRIV(ctx);

	if (md && md->hdr) {
		debug(50,"executing header commands");
		list_do(md->hdr, run_header_command, ctx);
	}
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{
	filter_cleanup(ctx);
	return SMFIS_CONTINUE; /* FIXME? */
}		

sfsistat
mlfi_close(SMFICTX *ctx)
{
	filter_cleanup(ctx);
	return SMFIS_CONTINUE; 
}

struct smfiDesc smfilter =
{
	"MailfromFilter",	
	SMFI_VERSION,	
	SMFIF_ADDHDRS|SMFIF_CHGHDRS,
	NULL,	/* connection info filter */
	NULL,	/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	NULL,	/* envelope recipient filter */
	NULL,	/* header filter */
	NULL,	/* end of header */
	NULL,	/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close,	/* connection cleanup */
};


/* Option handling.

   There are two classes of options: those that can be set using #pragma
   directive in the configuration file, and those that can be used only in
   command line.

   The latter are handled as usual. The former are processed in two stages:
   first, upon processing command line, their command line settings are
   verified and stored in struct option_cache below. Then, configuration
   file is parsed. Any valid `#pragma option' directives found there overwrite
   values in option_cache. Finally, process_options() is called that scans the
   cache and actually sets any options found there. */
   
void
set_ehlo(void *value)
{
	smtp_domain = value;
}

int
decode_debug(char *arg, int set)
{
	for (; *arg; arg++) {
		switch (*arg) {
		case 'd':
		case 'D':
			config_dump = 1;
			break;

		case 'y':
		case 'Y':
			config_ydebug = 1;
			break;

		case 'l':
		case 'L':
			config_ldebug = 1;
			break;

		default:
			if (isdigit(*arg)) {
				if (set)
					debug_level = debug_level * 10 +
						      *arg - '0';
			} else {
				mu_error("Unknown debug flag: %c", *arg);
				return 1;
			}
		}
	}
	return 0;
}

void
set_debug(void *value)
{
	decode_debug(value, 1);
}

void
set_positive_expire(void *value)
{
#ifdef USE_DBM
	positive_expire_interval = strtoul(value, NULL, 0);
#endif
}

void
set_negative_expire(void *value)
{
#ifdef USE_DBM
	negative_expire_interval = strtoul(value, NULL, 0);
#endif
}

void
set_expire(void *value)
{
#ifdef USE_DBM
	positive_expire_interval = negative_expire_interval =
		  strtoul(value, NULL, 0);
#endif
}

void
set_rates_expire(void *value)
{
#ifdef USE_DBM
	rates_expire_interval = strtoul(value, NULL, 0);
#endif
}

void
set_port(void *value)
{
	portspec = value;
}

void	
set_user(void *value)
{
	user = value;
}

void	
set_milter_timeout(void *value)
{
	if (smfi_settimeout(atoi(value)) == MI_FAILURE) {
		mu_error("Invalid milter timeout: %s", value);
		exit(EX_USAGE);
	}
}

void
set_pidfile(void *value)
{
	pidfile = value;
}

void
set_mailfrom(void *value)
{
	postmaster_email = value;
}

void
set_retry(void *value)
{
	io_attempts = strtoul(value, NULL, 0);
}

void
set_timeout(void *value)
{
	io_timeout = strtoul(value, NULL, 0);
}

static int
load_relay_file(void *item, void *data)
{
	read_domain_file(item);
	return 0;
}

void
set_relay(void *value)
{
	list_do(value, load_relay_file, NULL);
}

int
option_string(char *opt, void **pval, char *newval)
{
	if (*pval)
		free(*pval);
	*pval = strdup(newval);
	return 0;
}

int
option_number(char *opt, void **pval, char *newval)
{
	char *p;
	
	strtoul(newval, &p, 10);
	if (*p) {
		errno = EINVAL;
		return 1;
	}
	return option_string(opt, pval, newval);
}

int
option_debug(char *opt, void **pval, char *newval)
{
	if (decode_debug(newval, 0)) {
		errno = EINVAL;
		return 1;
	}
	return option_string(opt, pval, newval);
}

int
option_time(char *opt, void **pval, char *newval)
{
	return option_number(opt, pval, newval); /* FIXME */
}

int
option_email(char *opt, void **pval, char *newval)
{
	if (newval[0] == 0 || strchr(newval, '@'))
		return option_string(opt, pval, newval);
	errno = EINVAL;
	return 1;
}

int
option_pidfile(char *opt, void **pval, char *newval)
{
	if (newval[0] != '/') {
		mu_error("Invalid pidfile name: must be absolute");
		return 1;
	}
	return 0;
}

int
option_relay(char *opt, void **pval, char *newval)
{
	if (!*pval) 
		list_create((list_t*)pval);
	list_append(*pval, strdup(newval));
	return 0;
}

struct option_cache {
	char *name;
	void *value;
	int (*handler)(char *opt, void **pval, char *newval);
	void (*set)(void *value);
} option_cache[] = {
	{ "ehlo", NULL, option_string, set_ehlo },
	{ "debug", NULL, option_debug, set_debug },
	{ "expire-interval", NULL, option_time, set_expire },
	{ "positive-expire-interval", NULL, option_time, set_positive_expire },
	{ "negative-expire-interval", NULL, option_time, set_negative_expire },
	{ "rates-expire-interval", NULL, option_time, set_rates_expire },
	{ "port", NULL, option_string, set_port }, /* FIXME: option_port */
	{ "user", NULL, option_string, set_user },
	{ "milter-timeout", NULL, option_time, set_milter_timeout },
	{ "pidfile", NULL, option_pidfile, set_pidfile },
	{ "mailfrom", NULL, option_email, set_mailfrom },
	{ "retry", NULL, option_number, set_retry },
	{ "timeout", NULL, option_time, set_timeout },
	{ "relay", NULL, option_relay, set_relay },
	{ NULL }
};

static struct option_cache *
find_option(char *name)
{
	struct option_cache *p;

	for (p = option_cache; p->name; p++)
		if (strcmp(p->name, name) == 0)
			return p;
	return NULL;
}

int
set_option(char *name, char *value, int override)
{
	struct option_cache *p = find_option(name);
	if (!p) {
		errno = ENOENT;
		return 1;
	}
	if (!override && p->value)
		return 0;
	return p->handler(name, &p->value, value);
}

void
process_options()
{
	struct option_cache *p;

	for (p = option_cache; p->name; p++)
		if (p->value)
			p->set(p->value);
}


/* Command line parsing */

const char *program_version = "mailfromd (" PACKAGE_STRING ")";
static char doc[] = N_("mailfromd -- MAIL FROM milter checker");
static char args_doc[] = "";

#define OPTION_MILTER_TIMEOUT 256
#define OPTION_SHOW_DEFAULTS 257
#define OPTION_FOREGROUND 258
#define OPTION_PIDFILE 259
#define OPTION_DOMAIN_FILE 260
#define OPTION_TIMEOUT 261
#define OPTION_RETRY 262
#define OPTION_DELETE 263
#define OPTION_DAEMON 264
#define OPTION_LISTDB 265
#define OPTION_POSTMASTER_EMAIL 266
#define OPTION_EXPIRE 267
#define OPTION_TRACE 268
#define OPTION_PREDICT_NEXT 269
#define OPTION_SYSLOG 270

static struct argp_option options[] = {
#define GRP 0
	{ NULL, 0, NULL, 0,
	  "Operation modifiers", GRP },
#ifdef USE_DBM
	{ "delete", OPTION_DELETE, N_("DBNAME"), OPTION_ARG_OPTIONAL,
	  "Delete given entries from the cache", GRP+1 },
	{ "list", OPTION_LISTDB, N_("DBNAME"), OPTION_ARG_OPTIONAL,
	  "List cache (default) or rates database", GRP+1 },
	{ "expire", OPTION_EXPIRE, N_("DBNAME"), OPTION_ARG_OPTIONAL,
	  "Delete expired entries from the cache", GRP+1 },
	{ "predict", OPTION_PREDICT_NEXT, "RATE", 0,
	  "Use with --list-rate. Predict when the user will be able to send next message.", GRP+1 },
#endif
	{ "test", 't', NULL, 0,
	  "Run in test mode", GRP+1 },
	{ "show-defaults", OPTION_SHOW_DEFAULTS, NULL, 0,
	  "Show compilation defaults", GRP+1 },
	{ "daemon", OPTION_DAEMON, NULL, 0,
	  "Run in daemon mode (default)", GRP+1 },
#undef GRP	
#define GRP 10
	{ NULL, 0, NULL, 0,
	  "General options", GRP },
	{ "config-file", 'c', "FILE", 0,
	  "Read configuration from FILE", GRP+1 },
	{ "domain", 'D', "STRING", 0,
	  "Set default SMTP domain", GRP+1 },
	{ "ehlo", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
	{ "postmaster-email", OPTION_POSTMASTER_EMAIL, "EMAIL", 0,
	  "Set postmaster email address", GRP+1 },
	{ "mailfrom", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
	{ "port", 'p', "STRING", 0,
	  "Set communication socket", GRP+1 },
	{ "remove", 'r', NULL, 0,
	  "Force removing local socket file, if it already exists",
	  GRP+1 },
	{ "foreground", OPTION_FOREGROUND, NULL, 0,
	  "Stay in foreground", GRP+1 },
	{ "pidfile", OPTION_PIDFILE, "FILE", 0,
	  "Set pidfile name", GRP+1 },
	{ "user", 'u', "NAME", 0,
	  "Switch to this user privileges after startup", GRP+1 },
#ifdef USE_DBM
	{ "expire-interval", 'e', "NUMBER", 0,
	  "Set cache expiration interval to NUMBER seconds", GRP+1 },
#endif
#undef GRP

#undef GRP
#define GRP 30
	{ NULL, 0, NULL, 0,
	  "Timeout control", GRP },
	{ "milter-timeout", OPTION_MILTER_TIMEOUT, "NUMBER", 0,
	  "Set MTA connection timeout", GRP+1 },
	{ "timeout", OPTION_TIMEOUT, "NUMBER", 0,
	  "Set I/O operation timeout (seconds)", GRP+1 },
	{ "retry", OPTION_RETRY, "NUMBER", 0,
	  "Retry failed I/O operations NUMBER times", GRP+1 },
	{ "relayed-domain-file", OPTION_DOMAIN_FILE, "FILE", 0,
	  "Read relayed domains from FILE", GRP+1 },
	
#undef GRP
#define GRP 40
	{ NULL, 0, NULL, 0,
	  "Informational and debugging options", GRP },
	{ "transcript", 'X', NULL, 0,
	  "Enable transcript", GRP+1 },
	{ "trace", OPTION_TRACE, NULL, 0,
	  "Enable tracing configuration", GRP+1 },
	{ "debug", 'd', "LEVEL", 0,
	  "Set debugging level", GRP+1 },
	{ "stderr", 's', NULL, 0,
	  "Log to stderr", GRP+1 },
	{ "syslog", OPTION_SYSLOG, NULL, 0,
	  "Log to syslog (default)", GRP+1 },
#undef GRP
	{ NULL }
};


static int
select_mode(const char *arg, int mode1, int mode2)
{
	if (!arg || strcasecmp(arg, "cache") == 0) 
		mode = mode1;
	else if (strcasecmp(arg, "rates") == 0) 
		mode = mode2;
	else
		return 1;
	return 0;
}
	
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
	switch (key) {
	case 'c':
		config_file = arg;
		break;
		
	case 'D':
		set_option("ehlo", arg, 1);
		break;

	case 'X':
		do_transcript = 1;
		break;
		
	case 'd':
		set_option("debug", arg, 1);
		break;

#ifdef USE_DBM		
	case 'e':
		set_option("expire-interval", arg, 1);
		break;

	case OPTION_PREDICT_NEXT:
		if (convert_rate(arg, &predict_rate) == 0)
			predict_next_option = 1;
		break;
#endif
	case 'p':
		set_option("port", arg, 1);
		break;
		
	case 'r':
		force_remove = 1;
		break;

	case 's':
		log_to_stderr = 1;
		break;

	case 't':
		mode = MAILFROMD_TEST;
		log_to_stderr = isatty (2);
		need_config = 1;
		break;

	case 'u':
		set_option("user", arg, 1);
		break;

	case OPTION_DAEMON:
		mode = MAILFROMD_DAEMON;
		need_config = 1;
		break;
		
	case OPTION_DELETE:
		need_config = 0;
		log_to_stderr = isatty (2);
		if (select_mode(arg, MAILFROMD_DELETE, MAILFROMD_DELETE_RATEDB))
			argp_error(state, "Invalid database name: %s", arg);
		break;
		
	case OPTION_DOMAIN_FILE:
		set_option("relay", arg, 1);
		break;

	case OPTION_EXPIRE:
		need_config = 1;
		log_to_stderr = isatty (2);
		if (select_mode(arg, MAILFROMD_EXPIRE, MAILFROMD_EXPIRE_RATEDB))
			argp_error(state, "Invalid database name: %s", arg);
		break;
		
	case OPTION_FOREGROUND:
		foreground = 1;
		break;

	case OPTION_LISTDB:
		need_config = 0;
		log_to_stderr = isatty (2);
		if (select_mode(arg, MAILFROMD_LISTDB, MAILFROMD_LIST_RATEDB))
			argp_error(state, "Invalid database name: %s", arg);
		break;

	case OPTION_MILTER_TIMEOUT:
		set_option("milter-timeout", arg, 1);
		break;

	case OPTION_PIDFILE:
		set_option("pidfile", arg, 1);
		break;

	case OPTION_POSTMASTER_EMAIL:
		set_option("mailfrom", arg, 1);
		break;

	case OPTION_RETRY:
		set_option("retry", arg, 1);
		break;
		
	case OPTION_SHOW_DEFAULTS:
		mode = MAILFROMD_SHOW_DEFAULTS;
		need_config = 0;
		break;

	case OPTION_SYSLOG:
		log_to_stderr = 0;
		break;
		
	case OPTION_TIMEOUT:
		set_option("timeout", arg, 1);
		break;

	case OPTION_TRACE:
		do_trace = 1;
		break;
		
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static const char *capa[] = {
	"license",
	"logging",
	NULL
};

static struct argp argp = {
	options,
	parse_opt,
	args_doc,
	doc,
	NULL,
	NULL,
	NULL
};


/* Auxiliary functions */

/* Check whether local socket P exists. Act in accordance with the settings
   of --remove command line option. */
static void
check_local_portspec(char *p)
{
	struct stat st;
	
	if (stat(p, &st)) {
		if (errno == ENOENT)
			return; /* That's OK. The socket must not exist */
		mu_error("cannot stat file `%s': %s",
			 p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	} else if (!force_remove) {
		mu_error("File `%s' already exist. Remove it or use --remove option.", p);
		exit(EX_SOFTWARE);
	}

	if (unlink (p)) {
		mu_error("cannot unlink file `%s': %s", p, mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
}

/* Switch to the given UID/GID */
int
switch_to_privs(uid_t uid, gid_t gid)
{
	int rc = 0;
	gid_t emptygidset[1];

	/* Reset group permissions */
	emptygidset[0] = gid ? gid : getegid();
	if (geteuid() == 0 && setgroups(1, emptygidset)) {
		mu_error("setgroups(1, %lu) failed: %s",
		         (unsigned long) emptygidset[0],
			 mu_strerror(errno));
		rc = 1;
	}

	/* Switch to the user's gid. On some OSes the effective gid must
	   be reset first */

#if defined(HAVE_SETEGID)
	if ((rc = setegid(gid)) < 0)
		mu_error("setegid(%lu) failed: %s",
			(unsigned long) gid, mu_strerror(errno));
#elif defined(HAVE_SETREGID)
	if ((rc = setregid(gid, gid)) < 0)
		mu_error("setregid(%lu,%lu) failed: %s",
		         (unsigned long) gid, (unsigned long) gid,
			 mu_strerror(errno));
#elif defined(HAVE_SETRESGID)
	if ((rc = setresgid(gid, gid, gid)) < 0)
		mu_error("setresgid(%lu,%lu,%lu) failed: %s",
		         (unsigned long) gid,
		         (unsigned long) gid,
		         (unsigned long) gid,
			 mu_strerror(errno));
#endif

	if (rc == 0 && gid != 0) {
		if ((rc = setgid(gid)) < 0 && getegid() != gid) 
			mu_error("setgid(%lu) failed: %s",
				 (unsigned long) gid, mu_strerror(errno));
		if (rc == 0 && getegid() != gid) {
			mu_error("cannot set effective gid to %lu",
			         (unsigned long) gid);
			rc = 1;
		}
	}

	/* Now reset uid */
	if (rc == 0 && uid != 0) {
		uid_t euid;

		if (setuid(uid)
		    || geteuid() != uid
		    || (getuid() != uid
			&& (geteuid() == 0 || getuid() == 0))) {
			
#if defined(HAVE_SETREUID)
			if (geteuid() != uid) {
				if (setreuid(uid, -1) < 0) { 
					mu_error("setreuid(%lu,-1) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
				if (setuid(uid) < 0) {
					mu_error("second setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
			} else
#endif
				{
					mu_error("setuid(%lu) failed",
					         (unsigned long) uid,
						 mu_strerror(errno));
					rc = 1;
				}
		}
	
		euid = geteuid();
		if (uid != 0 && setuid(0) == 0) {
			mu_error("seteuid(0) succeeded when it should not");
			rc = 1;
		} else if (uid != euid && setuid(euid) == 0) {
			mu_error("cannot drop non-root setuid privileges");
			rc = 1;
		}

	}

	return rc;
}

/* Check whether pidfile NAME exists and if so, whether its PID is still
   active. Exit if it is. */
void
check_pidfile(char *name)
{
	unsigned long pid;
	FILE *fp = fopen(name, "r");
	if (!fp) {
		if (errno == ENOENT)
			return;
		mu_error("Cannot open pidfile `%s': %s",
			 name, mu_strerror(errno));
		exit(EX_TEMPFAIL);
	}
	if (fscanf(fp, "%lu", &pid) != 1) {
		mu_error("Cannot get pid from pidfile `%s'", name);
	} else {
		if (kill(pid, 0) == 0) {
			mu_error("%s appears to run with pid %lu. If it does not, remove `%s' and retry.",
				 program_invocation_short_name,
				 pid,
				 name);
			exit(EX_USAGE);
		}
	}
	fclose(fp);
	if (unlink(pidfile)) {
		mu_error("Cannot unlink pidfile `%s'",
			 name, mu_strerror(errno));
		exit(EX_USAGE);
	}
}

static char *license =
    "   This is mailfromd filter.\n"
    "   Copyright (C) 2005, Sergey Poznyakoff\n"
    "\n"
    "   ls is free software; you can redistribute it and/or modify\n"
    "   it under the terms of the GNU General Public License as published by\n"
    "   the Free Software Foundation; either version 2 of the License, or\n"
    "   (at your option) any later version.\n"
    "\n"
    "   GNU Mailutils is distributed in the hope that it will be useful,\n"
    "   but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "   GNU General Public License for more details.\n"
    "\n"
    "   You should have received a copy of the GNU General Public License along\n"
    "   with GNU Mailutils; if not, write to the Free Software Foundation,\n"
"   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n\n";

void
mailfromd_test(int argc, char **argv)
{
	int i;
	list_t dict = NULL;

	if (!argc)
		return;
	
	dict_init(&dict);
	for (i = 0; i < argc; i++) {
		char *p;
		if (p = strchr(argv[i], '=')) {
			*p++ = 0;
			dict_install(dict, argv[i], p);
		}
	}
	test_program(dict);
}

void
mailfromd_daemon()
{
	int rc;
	
	if (getuid() == 0) {
		if (!user) {
			mu_error("When running as root, --user option is mandatory.");
			exit(EX_USAGE);
		} else {
			struct passwd *pw = getpwnam(user);
			if (!pw) {
				mu_error("no such user: %s", user);
				exit(EX_SOFTWARE);
			}
			if (pw && switch_to_privs(pw->pw_uid, pw->pw_gid))
				exit(EX_SOFTWARE);
		}
	}

	if (!foreground)
		check_pidfile(pidfile);
	
	if (strncasecmp(portspec, "unix:", 5) == 0) {
		check_local_portspec(portspec + 5);
		umask(007);
	} else if (strncasecmp(portspec, "local:", 6) == 0) {
		check_local_portspec(portspec + 6);
		umask(007);
	}

	if (smfi_setconn(portspec) == MI_FAILURE) {
		mu_error("smfi_setconn: %s", mu_strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	if (smfi_register(smfilter) == MI_FAILURE) {
		mu_error("smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

	if (!foreground) {
		daemon(0, 0);
		daemon_create_pidfile(pidfile);
		rc = smfi_main();
		mu_error("OK. REMOVING PIDFILE");
		daemon_remove_pidfile();
	} else
		rc = smfi_main();
	exit(rc);
}

void
mailfromd_delete(int argc, char **argv)
{
	int i;
	for (i = 0; i < argc; i++)
		cache_delete(argv[i]);
}

void
mailfromd_listdb(int argc, char **argv)
{
	if (argc) {
		int i;
		for (i = 0; i < argc; i++)
			cache_list_item(argv[i]);
	} else
		cache_list_db();
}

void
mailfromd_delete_ratedb(int argc, char **argv)
{
	int i;
	for (i = 0; i < argc; i++)
		rate_delete(argv[i]);
}

void
mailfromd_list_ratedb(int argc, char **argv)
{
	if (argc) {
		int i;
		for (i = 0; i < argc; i++)
			rate_list_item(argv[i]);
	} else
		rate_list_db();
}

void
mailfromd_show_defaults()
{
	printf("method:              standard\n");
	printf("configuration file:  %s\n", DEFAULT_CONFIG_FILE);
	printf("user:                %s\n", DEFAULT_USER);
	printf("statedir:            %s\n", MAILFROMSTATEDIR);
	printf("socket:              %s\n", DEFAULT_SOCKET);
	printf("pidfile:             %s\n", DEFAULT_PIDFILE);
#ifdef USE_DBM
	printf("cache database:      %s\n", DEFAULT_DATABASE);
	printf("cache expiration:    %lu\n", DEFAULT_EXPIRE_INTERVAL);
	printf("rate database:       %s\n", DEFAULT_RATE_DATABASE);
	printf("rate expiration:     %lu\n", DEFAULT_EXPIRE_RATES_INTERVAL);
#else
	printf("cache database:      %s\n", "DISABLED");
#endif
}	

int
main(int argc, char **argv)
{
	int index;

	mu_license_text = license;
	mu_argp_init (program_version, PACKAGE_BUGREPORT);
	mu_argp_parse (&argp, &argc, &argv, 0, capa, &index, NULL);

	argv += index;
	argc -= index;

	if (need_config
	    && parse_config(config_file, config_ydebug, config_ldebug))
		exit(EX_USAGE);
	process_options();
	if (config_dump)
		print_config();

	/* Set up default values */
	if (!smtp_domain)
		mu_get_host_name(&smtp_domain);
	if (!log_to_stderr) {
		openlog(program_invocation_short_name,
			LOG_PID, log_facility);
		mu_error_set_print(syslog_error_printer);
	}
	
	switch (mode) {
	case MAILFROMD_DAEMON:
		mailfromd_daemon();
		break;
		
	case MAILFROMD_TEST:
		mailfromd_test(argc, argv);
		break;
		
	case MAILFROMD_DELETE:
		mailfromd_delete(argc, argv);
		break;

	case MAILFROMD_DELETE_RATEDB:
		mailfromd_delete_ratedb(argc, argv);
		break;

	case MAILFROMD_LISTDB:
		mailfromd_listdb(argc, argv);
		break;

	case MAILFROMD_LIST_RATEDB:
		mailfromd_list_ratedb(argc, argv);
		break;
		
	case MAILFROMD_EXPIRE:
		cache_expire_db();
		break;

	case MAILFROMD_EXPIRE_RATEDB:
		rate_expire_db();
		break;
		
	case MAILFROMD_SHOW_DEFAULTS:
		mailfromd_show_defaults();
		break;
	}

	exit(EX_OK);
}
